import logging
import ecs_logging

from uvicorn.logging import DefaultFormatter


class CustomLogger:
    logger_name = "logger"
    level = logging.DEBUG
    __logger: logging.Logger = None

    @classmethod
    def init(cls):
        cls.__logger = logging.getLogger(cls.logger_name)
        cls.__logger.setLevel(cls.level)
        cls.__logger.propagate = False

        formatter = DefaultFormatter(fmt="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                                     datefmt="%Y-%m-%d %H:%M:%S")

        console_handler = logging.StreamHandler()
        console_handler.setLevel(cls.level)
        console_handler.setFormatter(formatter)
        cls.__logger.addHandler(console_handler)

        handler = logging.StreamHandler()
        handler.setFormatter(ecs_logging.StdlibFormatter())
        cls.__logger.addHandler(handler)


    @classmethod
    @property
    def logger(cls):
        if cls.__logger is None:
            cls.init()
        return cls.__logger



import os
from urllib.parse import urlparse, urlunparse

from aias_common.access.configuration import S3StorageConfiguration
from aias_common.access.file import File
from aias_common.access.logger import Logger
from aias_common.access.storages.abstract import AbstractStorage
from fastapi_utilities import ttl_lru_cache

LOGGER = Logger.logger


class S3Storage(AbstractStorage):

    def get_configuration(self) -> S3StorageConfiguration:
        assert isinstance(self.storage_configuration, S3StorageConfiguration)
        return self.storage_configuration

    def get_storage_parameters(self):
        import boto3
        conf = self.get_configuration()
        if conf.is_anon_client:
            from botocore import UNSIGNED
            from botocore.client import Config

            client = boto3.client(
                "s3",
                region_name=conf.region,
                endpoint_url=conf.endpoint,
                config=Config(signature_version=UNSIGNED)
            )
        else:
            client = boto3.client(
                "s3",
                region_name=conf.region,
                endpoint_url=conf.endpoint,
                aws_access_key_id=conf.api_key.access_key,
                aws_secret_access_key=conf.api_key.secret_key,
            )

        return {"client": client}

    def supports(self, href: str):
        scheme = urlparse(href).scheme
        netloc = urlparse(href).netloc

        if scheme == "s3":
            return netloc == self.get_configuration().bucket
        elif scheme == "http" or scheme == "https":
            return f"{scheme}://{netloc}" == self.get_configuration().endpoint
        return False

    def exists(self, href: str):
        return self.is_dir(href) or self.is_file(href)

    def get_rasterio_session(self):
        import rasterio.session

        params = {}

        if self.get_configuration().is_anon_client:
            params["session"] = rasterio.session.AWSSession(
                aws_unsigned=True,
                region_name=self.get_configuration().region,
                endpoint_url=self.get_configuration().endpoint
            )
        else:
            params["session"] = rasterio.session.AWSSession(
                aws_access_key_id=self.get_configuration().api_key.access_key,
                aws_secret_access_key=self.get_configuration().api_key.secret_key,
                region_name=self.get_configuration().region,
                endpoint_url=self.get_configuration().endpoint
            )

        return params

    def __get_href_key(self, href: str):
        return urlparse(href).path.removeprefix(f"/{self.get_configuration().bucket}").removeprefix("/").removesuffix("/") 


    def pull(self, href: str, dst: str):
        import botocore.client

        super().pull(href, dst)

        client: botocore.client.BaseClient = self.get_storage_parameters()["client"]

        obj = client.get_object(Bucket=self.get_configuration().bucket, Key=self.__get_href_key(href))
        with open(dst, "wb") as f:
            for chunk in obj['Body'].iter_chunks(50 * 1024):
                f.write(chunk)

    def push(self, href: str, dst: str):
        super().push(href, dst)
        raise NotImplementedError("'push' method has not been implemented yet for s3 storage")

    @ttl_lru_cache(ttl=AbstractStorage.cache_tt, max_size=AbstractStorage.cache_size)
    def __head_object(self, href: str):
        conf = self.get_configuration()
        if self.__get_href_key(href):
            try:
                return self.get_storage_parameters()["client"].head_object(
                    Bucket=conf.bucket,
                    Key=self.__get_href_key(href))
            except:
                return None
        else:
            return None

    def is_file(self, href: str):
        import botocore.exceptions
        try:
            response = self.__head_object(href)
            return (response is not None) and response['ResponseMetadata']['HTTPStatusCode'] == 200
        except botocore.exceptions.ClientError:
            return False

    @ttl_lru_cache(ttl=AbstractStorage.cache_tt, max_size=AbstractStorage.cache_size)
    def __list_objects(self, href: str):
        conf = self.get_configuration()
        params = {
            "Bucket": conf.bucket,
            "Delimiter": "/",
            "MaxKeys": conf.max_objects
        }
        prefix = self.__get_href_key(href) + "/"
        if prefix != "/":
            params["Prefix"] = prefix
        return self.get_storage_parameters()["client"].list_objects_v2(**params)

    def is_dir(self, href: str) -> bool:
        objects = self.__list_objects(href)
        return (objects is not None) and (objects['KeyCount'] > 0 or len(objects.get('CommonPrefixes', [])) > 0)

    def get_file_size(self, href: str):
        response = self.__head_object(href)
        if response:
            return self.__head_object(href)['ContentLength']
        else:
            return None

    def __update_url__(self, source: str, path: str):
        url = urlparse(source)
        components = list(url[:])
        if len(components) == 5:
            components.append('')
        components[2] = os.path.join(self.get_configuration().bucket, path)
        return urlunparse(tuple(components))

    def listdir(self, source: str) -> list[File]:
        objects = self.__list_objects(source)
        files = []
        if objects.get("Contents"):
            files = list(map(lambda c: File(
                name=os.path.basename(c["Key"]),
                path=self.__update_url__(source, c["Key"]),
                is_dir=False,
                last_modification_date=c["LastModified"]), filter(lambda x: x.get("Key") != self.__get_href_key(source), objects["Contents"])))
        dirs = []
        if objects.get("CommonPrefixes"):
            dirs = list(map(lambda d: File(
                name=os.path.basename(d["Prefix"].removesuffix("/")),
                path=self.__update_url__(source, d["Prefix"]),
                is_dir=True), objects["CommonPrefixes"]))

        return files + dirs

    def get_last_modification_time(self, href: str):
        import botocore.exceptions
        try:
            response = self.__head_object(href)
            if response:
                return response['LastModified'].timestamp()
            return None
        except botocore.exceptions.ClientError:
            return None
        except botocore.exceptions.ParamValidationError:  # key LastModified does not exists for root
            return None

    def get_creation_time(self, href: str):
        # There is no difference in s3 between last update and creation date
        return self.get_last_modification_time(href)

    def makedir(self, href: str, strict=False):
        if strict:
            raise PermissionError("Creating a folder on a remote storage is not permitted")

    def clean(self, href: str):
        raise PermissionError("Deleting files on a remote storage is not permitted")

    def get_gdal_stream_options(self):
        config = self.get_configuration()

        params = {
            "AWS_VIRTUAL_HOSTING": "FALSE",
            # Before GDAL 3.11, http and https should not be in the endpoint adress
            "AWS_S3_ENDPOINT": config.endpoint.removeprefix("http://").removeprefix("https://")  # NOSONAR
        }

        if config.region != "auto":
            params["AWS_DEFAULT_REGION"] = config.region
        if config.is_anon_client:
            params["AWS_NO_SIGN_REQUEST"] = "YES"
        else:
            params["AWS_NO_SIGN_REQUEST"] = "NO"
            params["AWS_SECRET_ACCESS_KEY"] = config.api_key.secret_key
            params["AWS_ACCESS_KEY_ID"] = config.api_key.access_key

        # Not needed after GDAL 3.11
        if config.endpoint.startswith("http://"):  # NOSONAR
            params["AWS_HTTPS"] = "FALSE"
        return params

    def gdal_transform_href_vsi(self, href: str):
        config = self.get_configuration()

        if urlparse(href).scheme == "s3":
            href = href.replace("s3://", "/vsis3/")
        else:
            href = href.replace(config.endpoint, "/vsis3")

        return href



from typing import Callable, TypeAlias

from fastapi import status
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
from pydantic import BaseModel
from starlette.requests import Request

from aias_common.rest.exception import OGCException, RESTException

HandledExceptions: TypeAlias = RequestValidationError | OGCException


class ExceptionHandler(BaseModel, arbitrary_types_allowed=True):
    exception: type[Exception]
    handler: Callable[[Request, HandledExceptions], JSONResponse]


def validation_exception_handler(req: Request, exc: RequestValidationError):
    # Format the detail of the error message
    detail = ""
    for error in exc.errors():
        if len(error["loc"]) > 1:
            loc = error["loc"][1]
            for i in range(2, len(error["loc"])):
                if isinstance(error["loc"][i], str):
                    loc += f'.{error["loc"][i]}'
                elif isinstance(error["loc"][i], int):
                    loc += f'[{str(error["loc"][i])}]'
            detail += f'{loc}: {error["msg"]}\n'
        elif len(error["loc"]) > 0:
            detail += f'{error["loc"][0]}: {error["msg"]}\n'
    detail = detail[:-1]

    return JSONResponse(content=RESTException(
            type="bad request",
            status=status.HTTP_422_UNPROCESSABLE_ENTITY,
            title="validation error",
            detail=detail,
            instance=str(req.url)).model_dump(exclude_none=True, exclude_unset=True),
        status_code=status.HTTP_422_UNPROCESSABLE_ENTITY)


def server_error_handler(req: Request, exc: OGCException):
    return JSONResponse(content=RESTException(
            type=exc.type,
            status=exc.status,
            title=exc.title,
            detail=exc.detail,
            instance=str(req.url)).model_dump(exclude_none=True, exclude_unset=True),
        status_code=exc.status if exc.status is not None
        else status.HTTP_500_INTERNAL_SERVER_ERROR)


EXCEPTION_HANDLERS: list[ExceptionHandler] = [
    ExceptionHandler(exception=RequestValidationError,
                     handler=validation_exception_handler),
    ExceptionHandler(exception=OGCException,
                     handler=server_error_handler)
]



import time
from urllib.parse import urlparse

from aias_common.access.configuration import HttpStorageConfiguration
from aias_common.access.file import File
from aias_common.access.storages.abstract import AbstractStorage
from aias_common.access.storages.utils import (requests_exists, requests_get,
                                               requests_head)


class HttpStorage(AbstractStorage):

    def get_configuration(self) -> HttpStorageConfiguration:
        assert isinstance(self.storage_configuration, HttpStorageConfiguration)
        return self.storage_configuration

    def get_storage_parameters(self):
        return {"headers": self.get_configuration().headers}

    def supports(self, href: str):
        scheme = urlparse(href).scheme
        netloc = urlparse(href).netloc

        return scheme == self.get_configuration().type and netloc == self.get_configuration().domain

    def exists(self, href: str):
        return requests_exists(href, self.get_configuration().headers)

    def get_rasterio_session(self):
        # Might not work
        return {}

    def pull(self, href: str, dst: str):
        super().pull(href, dst)
        requests_get(href, dst, self.get_configuration().headers)

    def push(self, href: str, dst: str):
        super().push(href, dst)
        raise NotImplementedError("'push' method is not available for http storage")

    def is_file(self, href: str):
        return self.exists(href)

    def is_dir(self, href: str):
        return False

    def get_file_size(self, href: str):
        r = requests_head(href, self.get_configuration().headers)
        return r.headers.get("Content-Length")

    def listdir(self, href: str) -> list[File]:
        raise NotImplementedError(f"It is not possible to list the content of a directory with {self.get_configuration().type} protocol")

    def get_last_modification_time(self, href: str):
        r = requests_head(href, self.get_configuration().headers)
        return time.mktime(time.strptime(r.headers.get("Last-Modified"), "%a, %d %b %Y %H:%M:%S %Z"))

    def get_creation_time(self, href: str):
        # There is no difference in HTTP(S) between last update and creation date
        return self.get_last_modification_time(href)

    def makedir(self, href: str, strict=False):
        if strict:
            raise NotImplementedError(f"It is not possible to create the folder with {self.get_configuration().type} protocol")

    def clean(self, href: str):
        raise NotImplementedError(f"It is not possible to delete a file with {self.get_configuration().type} protocol")

    def get_gdal_stream_options(self):
        params = {}
        for (k, v) in self.get_configuration().headers.items():
            params[f"header.{k}"] = v

        return params

    def gdal_transform_href_vsi(self, href: str):
        return "/vsicurl/" + href



import os
from abc import ABC, abstractmethod
from urllib.parse import urlparse

from aias_common.access.configuration import AnyStorageConfiguration
from aias_common.access.file import File


class AbstractStorage(ABC):
    cache_tt = 60 * 5
    cache_size = 2048

    def __init__(self, storage_configuration: AnyStorageConfiguration):
        self.storage_configuration = storage_configuration

    @abstractmethod
    def get_configuration(self) -> AnyStorageConfiguration:
        """Returns the storage configuration

        Returns:
            AnyStorageConfiguration: storage configuration
        """
        ...

    @abstractmethod
    def get_storage_parameters(self) -> dict:
        """Based on the type of storage and its characteristics, gives storage-specific parameters to use to access data

        Args:
            href (str): Href of the file to consult
        """
        ...

    @abstractmethod
    def supports(self, href: str) -> bool:
        """Return whether the provided href can be handled by the storage.

        Args:
            href (str): Href of the file to consult

        Returns:
            bool: True if the storage can handle href, False otherwise
        """
        ...

    @abstractmethod
    def exists(self, href: str) -> bool:
        """Return whether the file given exists in the storage

        Args:
            href (str): Href of the file to consult

        Returns:
            bool: True if the file exists, False otherwise
        """
        ...

    @abstractmethod
    def get_rasterio_session(self) -> dict:
        """Return a rasterio Session and potential variables to access data remotely

        Args:
            href (str): Href od the file to stream

        Returns:
            dict
        """
        ...

    @abstractmethod
    def pull(self, href: str, dst: str):
        """Copy/Download the desired file from the file system to write it locally

        Args:
            href (str): File to fetch
            dst (str): Destination of the file
        """
        # Check that dst is local
        scheme = urlparse(dst).scheme
        if scheme != "" and scheme != "file":
            raise ValueError("Destination must be on the local filesystem")

    @abstractmethod
    def push(self, href: str, dst: str):
        """Copy/upload the desired file from local to write it on the file system

        Args:
            href (str): File to upload
            dst (str): Destination of the file
        """
        # Check that href is local
        scheme = urlparse(href).scheme
        if scheme != "" and scheme != "file":
            raise ValueError("Source file to upload must be on the local filesystem")

    @abstractmethod
    def is_file(self, href: str) -> bool:
        """Returns whether the specified href is a file

        Args:
            href(str): The href to test

        Returns:
            bool: Whether the input is a file
        """
        ...

    @abstractmethod
    def is_dir(self, href: str) -> bool:
        """Returns whether the specified href is a directory

        Args:
            href(str): The href to test

        Returns:
            bool: Whether the input is a directory
        """
        ...

    @abstractmethod
    def get_file_size(self, href: str) -> int | None:
        """Returns the size of the specified href

        Args:
            href(str): The href to examine
        """
        ...

    @abstractmethod
    def listdir(self, href: str) -> list[File]:
        """Returns the list of files and folders in the specified directory

        Args:
            href(str): The directory to examine
        """
        ...

    @abstractmethod
    def get_last_modification_time(self, href: str) -> float:
        """Returns the last modification time of the specified href

        Args:
            href(str): The href to examine

        Returns:
            float: the timestamp in seconds of last modification time
        """
        ...

    @abstractmethod
    def get_creation_time(self, href: str) -> float:
        """Returns the creation time of the specified href

        Args:
            href(str): The href to examine

        Returns:
            float: the timestamp in seconds of creation time
        """
        ...

    @abstractmethod
    def makedir(self, href: str, strict=False):
        """Create if needed (and possible) the specified dir

        Args:
            href(str): The href to the dir to create
            strict(bool): Whether to force the creation
        """
        ...

    def dirname(self, href: str):
        """Return the name of the directory containing the specified href

        Args:
            href(str): The href to examine
        """
        return os.path.dirname(href)

    @abstractmethod
    def clean(self, href: str):
        """If authorized, remove the given file

        Args:
            href(str): The href to delete
        """
        ...

    @abstractmethod
    def get_gdal_stream_options(self) -> dict:
        """Return the options to use to stream a GDAL file through its virtual file systems
        """
        ...

    @abstractmethod
    def gdal_transform_href_vsi(self, href: str):
        """Transform the archive's href into a format manageable by GDAL's virtual file systems

        Args:
            href(str): The href to examine
        """
        ...

    def get_gdal_src(self, href: str):
        """Return the archive's dataset through GDAL's virtual file systems

        Args:
            href(str): The href to examine
        """
        from osgeo import gdal
        from osgeo.gdalconst import GA_ReadOnly

        with gdal.config_options(self.get_gdal_stream_options()):
            src_ds = gdal.Open(self.gdal_transform_href_vsi(href), GA_ReadOnly)
        return src_ds

    def get_gdal_info(self, href: str, gdal_options):
        """Return the archive's info through GDAL's virtual file systems

        Args:
            href(str): The href to examine
        """
        from osgeo import gdal

        with gdal.config_options(self.get_gdal_stream_options()):
            return gdal.Info(self.gdal_transform_href_vsi(href), options=gdal_options)



import os
from urllib.parse import urlparse, urlunparse

from aias_common.access.configuration import GoogleStorageConfiguration
from aias_common.access.file import File
from aias_common.access.storages.abstract import AbstractStorage
from fastapi_utilities import ttl_lru_cache
from google.cloud.storage import Client
from google.oauth2 import service_account


class GoogleStorage(AbstractStorage):

    def get_configuration(self) -> GoogleStorageConfiguration:
        assert isinstance(self.storage_configuration, GoogleStorageConfiguration)
        return self.storage_configuration

    def get_storage_parameters(self):
        if self.get_configuration().is_anon_client:
            client = Client.create_anonymous_client()
        else:
            credentials = service_account.Credentials.from_service_account_info(self.get_configuration().api_key.model_dump())
            client = Client(project=self.get_configuration().api_key.project_id, credentials=credentials)

        return {"client": client}

    def supports(self, href: str):
        scheme = urlparse(href).scheme
        netloc = urlparse(href).netloc

        return scheme == "gs" and netloc == self.get_configuration().bucket

    def __get_bucket(self):
        client = self.get_storage_parameters()["client"]

        if self.get_configuration().is_anon_client:
            return client.bucket(self.get_configuration().bucket)
        else:
            # Try to retrieve a bucket (this makes an API request)
            return client.get_bucket(self.get_configuration().bucket)

    def __get_blob(self, href: str):
        bucket = self.__get_bucket()
        return bucket.get_blob(urlparse(href).path.lstrip("/") or "/")

    def __create_blob(self, href: str):
        bucket = self.__get_bucket()
        return bucket.blob(urlparse(href).path.lstrip("/") or "/")

    def exists(self, href: str):
        return self.is_file(href) or self.is_dir(href)

    def get_rasterio_session(self):
        import rasterio.session

        params = {
            "session": rasterio.session.GSSession(self.get_configuration().credentials_file),
            **self.__get_gdal_signed()
        }

        return params

    def __get_gdal_signed(self):
        params = {}
        if self.get_configuration().api_key is None:
            params["GS_NO_SIGN_REQUEST"] = "YES"
        else:
            params["GS_NO_SIGN_REQUEST"] = "NO"

        return params

    def pull(self, href: str, dst: str):
        super().pull(href, dst)

        blob = self.__get_blob(href)
        if blob is None:
            raise LookupError(f"Can't find {href}")

        blob.download_to_filename(dst)

    def push(self, href: str, dst: str):
        super().push(href, dst)

        blob = self.__create_blob(dst)
        if blob is None:
            raise LookupError(f"Can't create blob: {dst}")

        blob.upload_from_filename(href)

    @ttl_lru_cache(ttl=AbstractStorage.cache_tt, max_size=1024)
    def __list_blobs(self, source: str) -> list[File]:
        """
        Return a list of files contained in the specified folder, as well as subfolders
        """
        url = urlparse(source)
        path = url.path.removeprefix("/")
        blobs = self.__get_bucket().list_blobs(prefix=path, delimiter="/")
        files = list(filter(lambda f: f.path != path and f.name != "", map(lambda b: File(name=os.path.basename(b.name), path=self.__update_url__(source=source, path=b.name), is_dir=False, last_modification_date=b.updated, creation_date=b.time_created), blobs)))
        dirs = list(map(lambda b: File(name=os.path.basename(b.removesuffix("/")), path=self.__update_url__(source=source, path=b).removesuffix("/") + "/", is_dir=True), blobs.prefixes))
        return files + dirs

    def __update_url__(self, source: str, path: str):
        url = urlparse(source)
        components = list(url[:])
        if len(components) == 5:
            components.append('')
        components[2] = path
        return urlunparse(tuple(components))

    def is_file(self, href: str):
        files = self.__list_blobs(href)
        return len(list(filter(lambda f: f.path == href and not f.is_dir, files))) == 1

    def is_dir(self, href: str):
        files = self.__list_blobs(href.removesuffix("/") + "/")
        return len(list(filter(lambda f: f.path.removesuffix("/") != href.removesuffix("/"), files))) > 0

    def get_file_size(self, href: str):
        return self.__get_blob(href).size

    def listdir(self, href: str) -> list[File]:
        return self.__list_blobs(href.removesuffix("/") + "/")

    def get_last_modification_time(self, href: str):
        blob = self.__get_blob(href)
        if blob:
            mod_time = blob.updated
            return mod_time.timestamp() if mod_time is not None else 0
        return 0

    def get_creation_time(self, href: str):
        blob = self.__get_blob(href)
        if blob:
            creation_time = blob.time_created
            return creation_time.timestamp() if creation_time is not None else 0
        return 0

    def makedir(self, href: str, strict=False):
        if strict:
            raise PermissionError("Creating a folder on a remote storage is not permitted")

    def clean(self, href: str):
        raise PermissionError("Deleting files on a remote storage is not permitted")

    def get_gdal_stream_options(self):
        params = {
            **self.__get_gdal_signed()
        }

        if not self.get_configuration().is_anon_client:
            params["GS_OAUTH2_PRIVATE_KEY"] = self.get_configuration().api_key.private_key
            params["GS_OAUTH2_CLIENT_EMAIL"] = self.get_configuration().api_key.client_email

        return params

    def gdal_transform_href_vsi(self, href: str):
        return href.replace("gs://", "/vsigs/")



class URLUtil:
    @staticmethod
    def compose(*args) -> str:
        args = list(args)
        if len(args) == 1 and isinstance(args[0], list):
            args = args[0]
        return "/".join(args)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aias_common",
    version="0.6.29",
    author="Gisaïa",
    description="ARLAS AIAS common library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(where="src"),
    python_requires='>=3.10',
    package_dir={'': 'src'},
    install_requires=['ecs_logging', 'google-cloud-storage==2.5.0', 'pydantic==2.10.6', 'requests==2.31.0',
                      'smart_open==6.2.0', 'boto3==1.24.89', 'fastapi_utilities']
)
